from datetime import timedelta

from statsmodels.tsa.stattools import adfuller
from enodo import EnodoModule


class BaseModule:
    def __init__(self, dataset, params, series_name, siridb_client):
        """
        Start modelling a time serie
        :param series_name: name of the serie
        :param dataset: dataframe (Panda) with datapoints
        """
        raise NotImplementedError

    @classmethod
    def get_module_info(cls) -> EnodoModule:
        raise NotImplementedError


def adf_stationarity_test(timeseries):

    # Dickey-Fuller test:
    adf_test = adfuller(timeseries[1], autolag='AIC')

    p_value = adf_test[1]

    return p_value >= .05


def find_frequency(datetime_list):

    summed_freq = timedelta(seconds=0)
    summed_count = 0
    i = 1
    while i < len(datetime_list):
        if (i - 1) in datetime_list and i in datetime_list:
            summed_freq += abs(datetime_list[i - 1] - datetime_list[i])
            summed_count += 1
        i += 1

    return summed_freq / (summed_count)


# -----------------------------------------------------------------------------
# accept a dataframe, remove outliers, return cleaned data in a new dataframe
# see http://www.itl.nist.gov/div898/handbook/prc/section1/prc16.htm
# -----------------------------------------------------------------------------
def remove_outlier_in_df(df_in, col_name):
    q1 = df_in[col_name].quantile(0.05)
    q3 = df_in[col_name].quantile(0.95)
    iqr = q3 - q1  # Interquartile range
    fence_low = q1 - 1.5 * iqr
    fence_high = q3 + 1.5 * iqr
    df_out = df_in.loc[
        (df_in[col_name] >= fence_low) & (df_in[col_name] <= fence_high)]
    return df_out



import asyncio
import datetime
import os
import signal
import logging

from multiprocessing import Process, Queue

from enodo.client import Client
from enodo.protocol.package import *
from enodo.jobs import JOB_TYPE_FORECAST_SERIES, \
    JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES, JOB_TYPE_BASE_SERIES_ANALYSIS, \
    JOB_TYPE_STATIC_RULES
from enodo.protocol.packagedata import EnodoJobDataModel
from enodo.model.config.series import SeriesJobConfigModel

from .analyser import start_analysing
from .modules import module_load
from .lib.config import EnodoConfigParser
from .lib.logging import prepare_logger
from enodo.version import __version__ as VERSION


class Worker:

    def __init__(
            self, config_path, log_level, worker_name="Worker",
            worker_version="x.x.x"):
        self._loop = None
        self._log_level = log_level
        self._config = EnodoConfigParser()
        if config_path is not None and os.path.exists(config_path):
            self._config.read(config_path)
        self._client = None

        self._client_run_task = None
        self._updater_task = None
        self._result_queue = Queue()
        self._busy = False
        self._started_last_job = None
        self._max_job_duration = self._config['worker'][
            'max_job_duration']
        self._worker_process = None
        self._current_job = None
        self._running = True
        self._current_job_started_at = None
        self._modules = {}
        self._module_classes = {}

        prepare_logger(self._log_level)
        logging.info(f"Starting {worker_name} V{worker_version}, "
                     f"with lib V{VERSION}")

    async def _update_busy(self, busy, job_id=None):
        self._busy = busy
        self._current_job = job_id
        self._current_job_started_at = \
            datetime.datetime.now() if busy else None
        await self._client.send_message(busy, WORKER_UPDATE_BUSY)

    async def _send_refused(self):
        await self._client.send_message(None, WORKER_REFUSED)

    async def _send_shutdown(self):
        await self._client.send_message(None, CLIENT_SHUTDOWN)

    async def _check_for_update(self):
        while self._running:
            if not self._result_queue.empty():
                try:
                    result = self._result_queue.get()
                except Exception as e:
                    logging.error(
                        'Error while fetching item from result queue')
                    logging.debug(
                        f'Corresponding error: {str(e)}, '
                        f'exception class: {e.__class__.__name__}')
                else:
                    result['job_id'] = self._current_job
                    await self._send_update(result)
                    await self._update_busy(False)
            if self._busy and (
                datetime.datetime.now() - self._current_job_started_at)\
                    .total_seconds() >= int(self._max_job_duration):
                await self._cancel_job()
            await asyncio.sleep(2)

    async def _send_update(self, pkl):
        try:
            await self._client.send_message(pkl, WORKER_JOB_RESULT)
        except Exception as e:
            logging.error('Error while sending update to Hub')
            logging.debug(f'Corresponding error: {str(e)}, '
                          f'exception class: {e.__class__.__name__}')

    async def _receive_job(self, data):
        if self._busy:
            await self._send_refused()
        else:
            try:
                data = EnodoJobDataModel.unserialize(data)
                logging.info(
                    f'Received request for '
                    f'{data.get("job_config").get("job_type")} for series: '
                    f'"{data.get("series_name")}"')
            except Exception as e:
                logging.error(
                    'Error while unserializing incoming job data')
                logging.debug(
                    f'Corresponding error: {str(e)}',
                    f'exception class: {e.__class__.__name__}')
            await self._update_busy(True, data.get('job_id'))
            job_config = SeriesJobConfigModel(**data.get('job_config'))
            job_type = job_config.job_type

            if job_type in [JOB_TYPE_FORECAST_SERIES,
                            JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES,
                            JOB_TYPE_BASE_SERIES_ANALYSIS,
                            JOB_TYPE_STATIC_RULES]:
                module_name = job_config.module
                if not await self._check_support_job_and_module(
                        job_type, module_name):
                    await self._send_update(
                        {'error': 'Unsupported module for job_type',
                         'job_id': data.get('job_id'),
                         'name': data.get("series_name")})
                    await self._update_busy(False)
                    return
            else:
                await self._send_update(
                    {'error': 'Unsupported job_type',
                     'job_id': data.get('job_id'),
                     'name': data.get("series_name")})
                await self._update_busy(False)
                return

            try:

                self._worker_process = Process(
                    target=start_analysing,
                    args=(
                        self._result_queue, data,
                        self._config['siridb_data'],
                        self._config['siridb_output'],
                        self._module_classes
                    ))
                # Start the thread
                # self._worker_thread.daemon = True
                self._worker_process.start()
            except Exception as e:
                logging.error('Error while creating worker thread')
                logging.debug(
                    f'Corresponding error: {str(e)}, '
                    f'exception class: {e.__class__.__name__}')

    async def _check_support_job_and_module(self, job_type, module_name=None):
        modules = self._modules.values()
        if module_name is not None:
            modules = [self._modules.get(module_name)]
        for module in modules:
            if module.support_job_type(job_type):
                return True
        return False

    async def _cancel_job(self):
        try:
            self._worker_process.kill()
        except Exception as e:
            logging.error('Error while trying to cancel job')
            logging.debug(f'Corresponding error: {str(e)}, '
                          f'exception class: {e.__class__.__name__}')
        finally:
            await self._send_job_cancelled()

    async def _receive_to_cancel_job(self, data):
        job_id = data.get('job_id')
        if job_id is self._current_job:
            await self._cancel_job()

    async def _send_job_cancelled(self):
        await self._client.send_message({
            "job_id": self._current_job}, WORKER_JOB_CANCELLED)
        await self._update_busy(False)

    async def _add_handshake_data(self):
        serialized_module = list(self._modules.values())[0]
        return {'busy': self._busy,
                'module': serialized_module}

    def load_module(self, base_dir):
        # Get installed module
        modules = module_load(base_dir)
        logging.info("Loading installed analysis module:")
        for module_name, module_class in modules.items():
            self._modules[module_name] = module_class.get_module_info()
            self._module_classes[module_name] = module_class
            logging.info(f" - {module_name}")

    async def start_worker(self):
        if len(self._modules.keys()) < 1:
            logging.error("No module loaded")
            return

        self._loop = asyncio.get_running_loop()

        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            self._loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(
                    self.shutdown(s)))

        self._client = Client(
            self._loop, self._config['worker']['hub_hostname'],
            int(self._config['worker']['hub_port']),
            'worker', self._config['worker']
            ['internal_security_token'],
            heartbeat_interval=int(
                self._config['worker']['heartbeat_interval']),
            identity_file_path=".enodo_id", client_version=VERSION)

        await self._client.setup(cbs={
            WORKER_JOB: self._receive_job,
            WORKER_JOB_CANCEL: self._receive_to_cancel_job
        },
            handshake_cb=self._add_handshake_data)
        self._client_run_task = self._loop.create_task(
            self._client.run())
        self._updater_task = self._loop.create_task(
            self._check_for_update())

        await asyncio.gather(self._client_run_task)

    async def shutdown(self, s):
        self._running = False
        await self._send_shutdown()
        await self._client.close()

        """Cleanup tasks tied to the service's shutdown."""
        logging.info(f"Received exit signal {s.name}...")
        logging.info("Sending Hub that we are going down")
        tasks = [t for t in asyncio.all_tasks()
                 if t is not asyncio.current_task()]

        [task.cancel() for task in tasks]

        logging.info(f"Cancelling {len(tasks)} outstanding tasks")
        await asyncio.gather(*tasks)
        await asyncio.sleep(1)
        self._loop.stop()



import asyncio
from siridb.connector.lib.connection import SiriDBAsyncConnection

from .protocol import SiriDBServerProtocol


class PipeServer(SiriDBAsyncConnection):
    def __init__(self, pipe_name, on_data):
        self._pipe_name = pipe_name
        self._protocol = None
        self._server = None
        self._on_data_cb = on_data

    async def create(self, loop=None):
        loop = loop or asyncio.get_event_loop()

        self._server = await loop.create_unix_server(
            path=self._pipe_name,
            protocol_factory=lambda: SiriDBServerProtocol(self._on_data))

    def _on_data(self, data):
        """
        series names are returned as c strings (0 terminated)
        """
        data = {k.rstrip('\x00'): v for k, v in data.items()}
        self._on_data_cb(data)



import asyncio
import datetime
import os
import uuid

import qpack

from enodo.logging import client_logger as logger
from .protocol.package import *
from .version import __version__ as VERSION


class EnodoStreamReaderProtocol(asyncio.StreamReaderProtocol):

    def __init__(
            self, stream_reader, client_connected_cb=None, loop=None,
            client_lost_connection_cb=None):
        super(EnodoStreamReaderProtocol, self).__init__(
            stream_reader, client_connected_cb=None, loop=None)
        self._client_lost_connection_cb = client_lost_connection_cb

    def connection_lost(self, exc):
        super(EnodoStreamReaderProtocol, self).connection_lost(exc)
        if self._client_lost_connection_cb is not None:
            self._client_lost_connection_cb()


async def open_connection(host=None, port=None,
                          limit=asyncio.streams._DEFAULT_LIMIT,
                          client_lost_connection_cb=None):
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(limit=limit, loop=loop)
    protocol = EnodoStreamReaderProtocol(
        reader, loop=loop,
        client_lost_connection_cb=client_lost_connection_cb)
    transport, _ = await loop.create_connection(
        lambda: protocol, host, port)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
    return reader, writer


class Client:

    def __init__(self, loop, hostname, port, client_type, token,
                 identity_file_path=None, heartbeat_interval=5,
                 client_version=VERSION):
        self.loop = loop
        self._hostname = hostname
        self._port = port
        self._heartbeat_interval = heartbeat_interval
        self._client_type = client_type
        self._client_version = client_version
        self._id = uuid.uuid4().hex
        if identity_file_path is not None:
            enodo_id = self.read_enodo_id(identity_file_path)
            if enodo_id is None or enodo_id == "":
                self.write_enodo_id(identity_file_path, self._id)
            else:
                self._id = enodo_id

        self._token = token
        self._last_heartbeat_send = datetime.datetime.now()
        self._last_heartbeat_received = datetime.datetime.now()
        self._cbs = None
        self._handshake_data_cb = None
        self._reader = None
        self._writer = None
        self._running = True
        self._connected = False
        self._handsshaked = False
        self._read_task = None

    async def setup(self, cbs=None, handshake_cb=None):
        self._cbs = cbs
        if cbs is None:
            self._cbs = {}
        if handshake_cb is not None:
            self._handshake_data_cb = handshake_cb

    def read_enodo_id(self, path):
        enodo_id = os.getenv(f"ENODO_ID")
        if not (enodo_id is None or enodo_id == ""):
            return enodo_id
        if not os.path.exists(path):
            return None
        with open(path, 'r') as f:
            enodo_id = f.read()
        return enodo_id

    def write_enodo_id(self, path, enodo_id):
        with open(path, 'w') as f:
            f.write(enodo_id)
        return True

    async def _connect(self):
        while not self._connected:
            logger.info("Trying to connect")
            try:
                self._reader, self._writer = await open_connection(
                    self._hostname,
                    self._port,
                    client_lost_connection_cb=self.connection_lost)
            except Exception as e:
                logger.warning(f"Cannot connect, {str(e)}")
                logger.info("Retrying in 5")
                await asyncio.sleep(4)
            else:
                logger.info("Connected")
                self._connected = True
                await self._handshake()

    def connection_lost(self):
        logger.warning('Connection lost')
        self._connected = False
        self._writer.close()
        self._writer = None
        self._read_task.cancel()

    async def run(self):
        while self._running:
            if not self._connected:
                if self._handsshaked:
                    self._handsshaked = False
                if self._writer and not self._writer.is_closing():
                    self._writer.close()
                    await self._writer.wait_closed()
                await self._connect()
            elif self._handsshaked:
                diff = datetime.datetime.now() - self._last_heartbeat_send
                if diff.total_seconds() > int(
                        self._heartbeat_interval):
                    await self._send_heartbeat()
                diff = datetime.datetime.now() - self._last_heartbeat_received
                if diff.total_seconds() > int(
                        2*self._heartbeat_interval):
                    logger.error(
                        "Haven't received heartback response from hub")
                    self._connected = False
            await asyncio.sleep(1)

    async def close(self):
        logger.info('Closing the socket')
        self._running = False
        if self._read_task:
            self._read_task.cancel()
        if self._writer:
            self._writer.close()

    async def _read_from_socket(self):
        while self._running:
            if not self._connected:
                await asyncio.sleep(1)
                continue

            packet_type, packet_id, data = await read_packet(self._reader)
            if data is False:
                self._connected = False
                continue

            if len(data):
                data = qpack.unpackb(data, decode='utf-8')

            if packet_type == 0:
                logger.warning("Connection lost, trying to reconnect")
                self._connected = False
                try:
                    await self.setup(self._cbs)
                except Exception as e:
                    logger.error('Error while trying to setup client')
                    logger.debug(f'Correspondig error: {str(e)}')
                    await asyncio.sleep(5)
            elif packet_type == HANDSHAKE_OK:
                logger.info(f'Hands shaked with hub')
                self._last_heartbeat_received = datetime.datetime.now()
                self._handsshaked = True
            elif packet_type == HANDSHAKE_FAIL:
                logger.error(f'Hub does not want to shake hands')
            elif packet_type == HEARTBEAT:
                logger.debug(f'Heartbeat back from hub')
                self._last_heartbeat_received = datetime.datetime.now()
            elif packet_type == RESPONSE_OK:
                logger.debug(f'Hub received update correctly')
            elif packet_type == UNKNOWN_CLIENT:
                logger.error(f'Hub does not recognize us')
                await self._handshake()
            else:
                if packet_type in self._cbs.keys():
                    await self._cbs.get(packet_type)(data)
                else:
                    logger.error(
                        f'Message type not implemented: {packet_type}')
            await asyncio.sleep(1)

    async def _send_message(self, length, message_type, data):
        header = create_header(length, message_type)

        logger.debug(f"Sending type: {message_type}")
        self._writer.write(header + data)
        try:
            await self._writer.drain()
        except Exception as e:
            self._connected = False

    async def send_message(self, body, message_type, use_qpack=True):
        if not self._connected:
            return False
        if use_qpack:
            body = qpack.packb(body)
        await self._send_message(len(body), message_type, body)

    async def _handshake(self):
        self._read_task = asyncio.Task(self._read_from_socket())
        data = {
            'client_id': str(self._id),
            'client_type': self._client_type,
            'token': self._token,
            'lib_version': self._client_version}
        if self._handshake_data_cb is not None:
            handshake_data = await self._handshake_data_cb()
            data = {**data, **handshake_data}
        data = qpack.packb(data)
        await self._send_message(len(data), HANDSHAKE, data)
        self._last_heartbeat_send = datetime.datetime.now()

    async def _send_heartbeat(self):
        logger.debug('Sending heartbeat to hub')
        id_encoded = qpack.packb(self._id)
        await self._send_message(len(id_encoded), HEARTBEAT, id_encoded)
        self._last_heartbeat_send = datetime.datetime.now()



import os
import importlib


def module_load(base_path):
    modules = {}

    bootstrap_file_path = os.path.join(
        base_path, "module",
        "bootstrap.py")
    if os.path.isfile(bootstrap_file_path):
        mod = importlib.import_module(
            f"module")
        c = mod.get_module_class()
        module_info = c.get_module_info()
        modules[module_info.name] = c

    return modules



import logging

AVAILABLE_LOG_LEVELS = {
    'error': logging.ERROR, 'warning': logging.WARNING,
    'info': logging.INFO, 'debug': logging.DEBUG}


def prepare_logger(log_level_label):
    log_level = AVAILABLE_LOG_LEVELS.get(log_level_label)
    logger = logging.getLogger()
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        fmt='[%(levelname)1.1s %(asctime)s worker] ' +
            '%(message)s',
        datefmt='%y%m%d %H:%M:%S',
        style='%')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logging.getLogger('enodo.client').setLevel(log_level)



import asyncio
import logging

from .package import Package


class SiriDBServerProtocol(asyncio.Protocol):

    def __init__(self, on_package_received):
        self._buffered_data = bytearray()
        self._data_package = None
        self._on_package_received = on_package_received

    def data_received(self, data):
        '''
        override asyncio.Protocol
        '''
        self._buffered_data.extend(data)
        while self._buffered_data:
            size = len(self._buffered_data)
            if self._data_package is None:
                if size < Package.struct_datapackage.size:
                    return None
                self._data_package = Package(self._buffered_data)
            if size < self._data_package.length:
                return None
            try:
                self._data_package.extract_data_from(self._buffered_data)
            except KeyError as e:
                logging.error('Unsupported package received: {}'.format(e))
            except Exception as e:
                logging.exception(e)
                # empty the byte-array to recover from this error
                self._buffered_data.clear()
            else:
                self._on_package_received(self._data_package.data)
            self._data_package = None



import logging

from siridb.connector import SiriDBClient
from siridb.connector.lib.exceptions import QueryError, InsertError, \
    ServerError, PoolError, AuthenticationError, UserAuthError


def config_equals(a, b):
    if a.get('user') != b.get('user'):
        return False
    if a.get('password') != b.get('password'):
        return False
    if a.get('database') != b.get('database'):
        return False
    if a.get('host') != b.get('host'):
        return False

    return True


class SiriDB:
    siri = None
    siridb_connected = False
    siridb_status = ""

    def __init__(self, siridb_user, siridb_password, siridb_db,
                 siridb_host, siridb_port):
        self.siri = SiriDBClient(
            username=siridb_user,
            password=siridb_password,
            dbname=siridb_db,
            # Multiple connections are supported
            hostlist=[(siridb_host, siridb_port)],
            keepalive=True)

    # @classmethod
    async def query_series_datapoint_count(self, series_name):
        await self.siri.connect()
        count = None
        try:
            result = await self.siri.query(
                f'select count() from "{series_name}"')
        except (QueryError, InsertError, ServerError,
                PoolError, AuthenticationError, UserAuthError) as e:
            logging.error('Connection problem with SiriDB server')
        else:
            count = result.get(series_name, [])[0][1]
        self.siri.close()
        return count

    # @classmethod
    async def query_series_data(self, series_name, max_n_points,
                                selector="*"):

        query = (f'select {selector} from "{series_name}" '
                 f'tail {max_n_points}')
        await self.siri.connect()
        result = None
        try:
            result = await self.siri.query(query)
        except (QueryError, InsertError, ServerError, PoolError,
                AuthenticationError, UserAuthError) as e:
            logging.error('Connection problem with SiriDB server')
        self.siri.close()
        return result

    async def run_query(self, query):
        await self.siri.connect()
        result = None
        try:
            result = await self.siri.query(query)
        except (QueryError, InsertError, ServerError, PoolError,
                AuthenticationError, UserAuthError) as e:
            logging.error('Connection problem with SiriDB server')
        self.siri.close()
        return result

    async def test_connection(self):
        try:
            await self.siri.connect()
        except Exception:
            return "Cannot connect", False

        try:
            await self.siri.query(f'show dbname')
        except (QueryError, InsertError, ServerError, PoolError,
                AuthenticationError, UserAuthError) as e:
            self.siri.close()
            return repr(e), False
        else:
            self.siri.close()
            return "", True
